#ifndef IMAPFILTER_H
#define IMAPFILTER_H

#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>

#ifndef DATA_H
#include "data.h"
#endif

/* Program's version number. */
#define IMAPFILTER_VERSION		"0.8.4"

/* Error codes returned by functions. */
#define ERROR_SIGNAL			1
#define ERROR_TRIVIAL			2
#define ERROR_FILE_OPEN			3
#define ERROR_LOCKFILE			4
#define ERROR_CONFIG_PARSE		5
#define ERROR_MEMORY_ALLOCATION		6
#define ERROR_SETUID			7
#define ERROR_TERMIO			8
#define ERROR_NETWORK			9
#define ERROR_SSL			10
#define ERROR_PASSPHRASE		11
#define ERROR_ENCRYPT			12
#define ERROR_DECRYPT			13
#define ERROR_FORK			14
#define ERROR_UNDEFINED			15

/* Flags that control the program's execution options. */
#define OPTION_DETAILS_QUIET		0x0001
#define OPTION_DETAILS_NORMAL		0x0002
#define OPTION_DETAILS_VERBOSE		0x0004
#define OPTION_DETAILS_CLEAR		0xfff8
#define OPTION_ERRORS			0x0008
#define OPTION_EXPUNGE			0x0010
#define OPTION_HEADERS			0x0020
#define OPTION_NAMESPACE		0x0040
#define OPTION_SUBSCRIBE		0x0080
#define OPTION_WARNING			0x0100
#define OPTION_PASSWORD_EDITOR		0x0200
#define OPTION_DAEMON_MODE		0x0400

/* Other flags. */
#define FLAG_TTY_MODIFIED		0x01
#define FLAG_BLANK_PASSWORD		0x02

/* Capabilities of mail server. */
#define CAPABILITY_NONE                 0x00
#define CAPABILITY_NAMESPACE            0x01

/* Flags for logger. */
#define LOG_WRITE			0
#define LOG_SERVER			1
#define LOG_USERNAME			2
#define LOG_MAILBOX			3
#define LOG_FILTER			4
#define LOG_ACTION			5
#define LOG_DESTINATION_MAILBOX		6

/* Buffer size of the IMAP command. */
#define SMALL_CMD			64
#define MEDIUM_CMD			512
#define BIG_CMD				32768

/* Store IMAP command type may replace,add or remove flags. */
#define STORE_FLAG_REPLACE		1
#define STORE_FLAG_ADD			2
#define STORE_FLAG_REMOVE		3

/* Buffer size of server's response. */
#define RESPONSE_BUF			(4096 + 1)
#define RESULT_BUF			512

/* Response codes. */
#define RESPONSE_NONE			-2
#define RESPONSE_BAD			-1
#define RESPONSE_OK			0
#define RESPONSE_NO			1
#define RESPONSE_BYE			2
#define RESPONSE_PREAUTH		3
#define RESPONSE_READONLY		4
#define RESPONSE_TRYCREATE		5

/* Size virtual buffer. */
#define VIRTUAL_BUF			4096

/* Length of mailbox namespace prefix. */
#define NAMESPACE_PREFIX_LEN            64

/* Encryption and decryption buffers. */
#define ENCRYPTION_BUF			1024
#define DECRYPTION_BUF			4096

/* Buffer to save flags of the store IMAP command. */
#define STORE_FLAGS_BUF			64

/* Maximum passwords the password editor can handle. */
#define EDITOR_PASSWORDS_MAX		64


#define min(A, B)			((A) < (B) ? (A) : (B))
#define plural(A)			((A) == 1 ? "" : "s")


/* Secure memory information. */
typedef struct secmem {
    void *buf;			/* Allocated memory buffer. */
    size_t size;		/* Size of the buffer. */
    struct secmem *prev, *next;	/* Previous/next node of doubly linked list. */
} secmem_t;


/* Namespace of IMAP mailboxes. */
typedef struct namesp {
    char prefix[NAMESPACE_PREFIX_LEN];
    char delim;
} namesp_t;


/*	file.c		*/
int read_config(char *cfg);
int parse_config(FILE * fd);
void set_options(char *line, regmatch_t * match);

int read_passwords(void);
int parse_passwords(FILE * fd);
int store_passwords(account_t * accts[]);

int create_homedir(void);
int exists_file(char *fname);
int exists_dir(char *fname);
int create_file(char *fname, mode_t mode);
#ifdef CHECK_PERMISSIONS
int check_file_perms(char *fname, mode_t mode);
int check_dir_perms(char *dname, mode_t mode);
#endif

/*	imapfilter.c	*/
void usage(void);
void version(void);

/*	imap.c		*/
unsigned int send_command(int *sock, char *cmd);
#ifdef DEBUG
int imap_noop(int *sock);
#endif
int imap_capability(int *sock);
int imap_namespace(int *sock);
int imap_logout(int *sock);
int imap_login(int *sock, char *user, char *pass);
/* int imap_list(int *sock, char *refer, char *mbox); */
int imap_subscribe(int *sock, char *mbox);
int imap_examine(int *sock, char *mbox);
int imap_select(int *sock, char *mbox);
int imap_status(int *sock, char *mbox, char *items);
int imap_create(int *sock, char *mbox);
int imap_search(int *sock, char *search);
int imap_fetch(int *sock, char *mesg, char *items);
int imap_store(int *sock, char *mesg, unsigned int mode, char *flags);
int imap_copy(int *sock, char *mesg, char *mbox);
int imap_append(int *sock, char *mbox, unsigned int size);
int imap_close(int *sock);
int imap_expunge(int *sock);

/*	lock.c		*/
void lockfile_create(void);
void lockfile_check(void);
pid_t lockfile_pid(void);
int lockfile_remove(void);
void kill_imapfilter(void);

/*	log.c		*/
void info(const char *info,...);
void verbose(const char *info,...);
void error(const char *errmsg,...);
void fatal(unsigned int errnum, const char *fatal,...);
void catch_signals(void);
void signal_handler(int sig);
int open_logfile(void);
int create_logfile(void);
int close_logfile(void);
void log_info(int flag, void *ptr);
char *get_time(void);

/*	memory.c	*/
void *xmalloc(size_t size);
void *xrealloc(void *ptr, size_t size);
void xfree(void *ptr);
char *xstrdup(const char *s);

void *smalloc(size_t size);
void *srealloc(void *ptr, size_t size);
void sfree(void *ptr);
char *sstrdup(const char *s);
void secmem_append(secmem_t * node);
secmem_t *secmem_find(void *ptr);
void secmem_remove(secmem_t * node);
void secmem_clear(void);
#ifdef MEMORY_LOCK
void secmem_lock(void);
#endif
void corefile_disable(void);

/*	misc.c		*/
char *strcasestr(const char *haystack, const char *needle);
char *ultostr(unsigned long int num, int base);
char *xstrncpy(char *dest, const char *src, size_t size);

/*	passwd.c	*/
void get_password(char *passwd, size_t pwlen);
#ifdef ENCRYPTED_PASSWORDS
int encrypt_passwords(FILE * fd, account_t * accts[]);
int decrypt_passwords(unsigned char **buf, FILE * fd);
void password_editor(void);
#endif

/*	request.c	*/
int test(int *sock);
int check_capabilities(int *sock);
int check_namespace(int *sock, namesp_t * nsp);
int login(int *sock, char *user, char *pass);
int check_mailbox(int *sock, char *mbox, namesp_t * nsp);
int select_mailbox(int *sock, char *mbox, namesp_t * nsp);
int mailbox_status(int *sock, char *mbox, namesp_t * nsp);
int close_mailbox(int *sock);
int logout(int *sock);

int apply_filters(filter_t ** filters);
int match_filter(filter_t * filter, char **mesgs);

void empty_fifo(mask_t ** mfifo);
void queue_fifo(mask_t ** mfifo, mask_t * mask);
mask_t *dequeue_fifo(mask_t ** mfifo);

char *generate_filter_and(mask_t * mask, unsigned int masknum,
			   unsigned int masklen);
char *generate_filter_or(mask_t * mask, unsigned int masknum,
			  unsigned int masklen);

int apply_action(char *mesgs, unsigned int *type, account_t * raccount,
		  char *destmbox, unsigned int *msgflags, char *args);
int action_delete(char *mesgs, char *args);
int action_copy(char *mesgs, char *destmbox, char *args);
int action_move(char *mesgs, char *destmbox, char *args);
int action_rcopy(char *mesgs, account_t * destacc, char *destmbox, char *args);
int action_rmove(char *mesgs, account_t * destacc, char *destmbox, char *args);
int action_flag(char *mesgs, unsigned int *type, unsigned int *msgflags, char *args);
int action_list(char *mesgs, char *args);

unsigned int count_messages(char *mesgs);
char *convert_messages(char *mesgs);


/* 	response.c	*/
void receive_response(int *sock, char *buf);
int server_response(int *sock, unsigned int tag);
void bye_response(char *buf);
int greeting_response(int *sock);
int logout_response(int *sock, unsigned int tag);
int capability_response(int *sock, unsigned int tag);
int namespace_response(int *sock, unsigned int tag, namesp_t * nsp);
int status_response(int *sock, unsigned int tag, char *mbox);
int select_response(int *sock, unsigned int tag);
int search_response(int *sock, unsigned int tag, char **mesgs);
int fetch_response(int *sock, unsigned int tag, int reset, char *fetch);
int fetchsize_response(int *sock, unsigned int *size, unsigned int tag);
int append_response(int *sock, unsigned int tag);
int copy_response(int *sock, unsigned int tag);
int analyze_response(char *buf);
void init_vbuf(void);
void reset_vbuf(void);
void check_vbuf(size_t n);

/*	socket.c	*/
int init_connection(int *sock, char *serv, unsigned short int port,
		     unsigned int protocol);
int ssl_init(int *sock, unsigned int protocol);
int close_connection(int *sock);
int socket_read(int *sock, char *buf);
int socket_write(int *sock, char *data);

/*	tty.c		*/
int tty_store(void);
int tty_disable_echo(void);
int tty_restore(void);

#endif				/* IMAPFILTER_H */
