#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>

#include "config.h"
#include "imapfilter.h"

#ifdef SSL_TLS
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/x509.h>
#endif


long timeout = -1;		/* Server non-response timeout in seconds. */

int sockpri = -1;		/* Main socket used mostly. */
int sockaux = -1;		/* Auxiliary socked used when another
				   connection is needed. */
#ifdef SSL_TLS
static SSL *sslpri = NULL;
static SSL *sslaux = NULL;
#endif


/*
 * Connect to mail server.
 */
int init_connection(int *sock, char *serv, unsigned short int port,
		     unsigned int protocol)
{
    struct sockaddr_in sa;
    struct hostent *he;
#ifdef SSL_TLS
    SSL **ssl = (sock == &sockpri ? &sslpri : &sslaux);
#else
    if (protocol != SSL_DISABLED) {
	error("imapfilter: SSL not supported by this build\n");
	return ERROR_SSL;
    }
#endif

    memset((char *)&sa, 0, sizeof(struct sockaddr_in));

    *sock = socket(PF_INET, SOCK_STREAM, 0);

    if (*sock < 0) {
	error("imapfilter: create socket; %s\n", strerror(errno));
	return ERROR_NETWORK;
    }
    if (!(he = gethostbyname(serv))) {
	error("imapfilter: get network host entry of %s; %s\n", serv,
	      strerror(errno));
	close_connection(sock);
	return ERROR_NETWORK;
    }
    sa.sin_family = AF_INET;
    sa.sin_port = htons(port);
    sa.sin_addr = *(struct in_addr *) he->h_addr;

    xstrncpy(serv, he->h_name, SERVER_LEN - 1);

    if (connect(*sock, (struct sockaddr *) & sa, sizeof(struct sockaddr))) {
	error("imapfilter: initiating connection to %s; %s\n", serv,
	      strerror(errno));
	close_connection(sock);
	return ERROR_NETWORK;
    }
    log_info(LOG_SERVER, serv);

#ifdef SSL_TLS
    if (protocol != SSL_DISABLED)
	if (!ssl_init(sock, protocol)) {
	    if (sock == &sockpri)
		info("Connected to %s using %s.\n", serv,
		     SSL_get_cipher(*ssl));
	    return 0;
	} else
	    return ERROR_SSL;
    else
	*ssl = NULL;
#endif

    if (sock == &sockpri)
	info("Connected to %s.\n", serv);

    return 0;
}


#ifdef SSL_TLS
/*
 * Initialize Secure Socket Layer connection.
 */
int ssl_init(int *sock, unsigned int protocol)
{
    int e;
    unsigned int i, n = 0;
    SSL_CTX *ctx;
    SSL_METHOD *method = NULL;
    X509 *cert;
    char *c;
    EVP_MD *evp;
    unsigned char digest[EVP_MAX_MD_SIZE];
    SSL **ssl = (sock == &sockpri ? &sslpri : &sslaux);

    SSL_library_init();

    switch (protocol) {
    case SSL_SSL_V2:
	method = SSLv2_client_method();
	break;
    case SSL_SSL_V3:
	method = SSLv3_client_method();
	break;
    case SSL_TLS_V1:
	method = TLSv1_client_method();
	break;
    }

    if (!(ctx = SSL_CTX_new(method)))
	return ERROR_SSL;

    if (!(*ssl = SSL_new(ctx)))
	return ERROR_SSL;

    SSL_set_fd(*ssl, *sock);

    e = SSL_connect(*ssl);

    if (e <= 0) {
	SSL_get_error(*ssl, e);
	error("imapfilter: initiating SSL connection; %s",
	      ERR_error_string(ERR_get_error(), NULL));
	return ERROR_SSL;
    }
    /* Get server's certificate. */
    if (!(cert = SSL_get_peer_certificate(*ssl)))
	return ERROR_SSL;

    if (!(c = X509_NAME_oneline(X509_get_subject_name(cert), 0, 0)))
	return ERROR_SSL;
    verbose("Server certificate subject: %s\n", c);
    xfree(c);

    if (!(c = X509_NAME_oneline(X509_get_issuer_name(cert), 0, 0)))
	return ERROR_SSL;
    verbose("Server certificate issuer: %s\n", c);
    xfree(c);

    if (!(evp = EVP_md5()))
	return ERROR_SSL;
    if (!(X509_digest(cert, evp, digest, &n)))
	return ERROR_SSL;

    verbose("Server key fingerprint: ");
    for (i = 0; i < n; i++)
	verbose(i != n - 1 ? "%02X:" : "%02X\n", digest[i]);

    X509_free(cert);
    SSL_CTX_free(ctx);

    return 0;
}

#endif


/*
 * Disconnect from mail server.
 */
int close_connection(int *sock)
{
#ifdef SSL_TLS
    SSL **ssl = (sock == &sockpri ? &sslpri : &sslaux);

    if (*ssl) {
	SSL_shutdown(*ssl);
	SSL_free(*ssl);
	*ssl = NULL;
    }
#endif

    if (*sock >= 0 && close(*sock)) {
	error("imapfilter: closing socket; %s\n", strerror(errno));
	return ERROR_NETWORK;
    } else {
	*sock = -1;
	return 0;
    }
}


/*
 * Read data from socket.
 */
int socket_read(int *sock, char *buf)
{
    int f, e, s;
    fd_set fds;
    struct timeval tv;
    struct timeval *tvp = NULL;
#ifdef SSL_TLS
    SSL **ssl = (sock == &sockpri ? &sslpri : &sslaux);
#endif

    e = 0;
    s = 1;

    memset(buf, 0, RESPONSE_BUF);

    if (timeout >= 0) {
	tv.tv_sec = timeout;
	tv.tv_usec = 0;
	tvp = &tv;
    }
    f = fcntl(*sock, F_GETFL, 0);
    fcntl(*sock, F_SETFL, f | O_NONBLOCK);

    FD_ZERO(&fds);
    FD_SET(*sock, &fds);

#ifdef SSL_TLS
    if (*ssl) {

	for (;;) {
	    if (SSL_pending(*ssl)
		|| ((s = select(*sock + 1, &fds, NULL, NULL, tvp)) > 0
		    && FD_ISSET(*sock, &fds)))
		e = SSL_read(*ssl, buf, RESPONSE_BUF - 1);

	    if (e > 0)
		break;

	    switch (SSL_get_error(*ssl, e)) {
	    case SSL_ERROR_WANT_READ:
	    case SSL_ERROR_WANT_WRITE:
		continue;
	    case SSL_ERROR_SYSCALL:
	    case SSL_ERROR_SSL:
		fatal(ERROR_NETWORK, "imapfilter: reading data; %s",
		      ERR_error_string(ERR_get_error(), NULL));
	    default:
		fatal(ERROR_NETWORK,
		    "imapfilter: undefined ssl error while reading data\n");
	    }
	}
    } else
#endif
	if ((s = select(*sock + 1, &fds, NULL, NULL, tvp)) > 0
	    && FD_ISSET(*sock, &fds))
	e = read(*sock, buf, RESPONSE_BUF - 1);

    if (e == -1)
	fatal(ERROR_NETWORK, "imapfilter: reading data; %s",
	      strerror(errno));

    fcntl(*sock, F_SETFL, f);

    if (s == -1)
	fatal(ERROR_NETWORK, "imapfilter: waiting to read from socket; %s\n",
	      strerror(errno));
    else if (!s)
	fatal(ERROR_NETWORK,
	      "imapfilter: timeout period expired while waiting to read "
	      "data\n");

    return 0;
}


/*
 * Write data to socket.
 */
int socket_write(int *sock, char *data)
{
    int f, e, s;
    fd_set fds;
    struct timeval tv;
    struct timeval *tvp = NULL;
#ifdef SSL_TLS
    SSL **ssl = (sock == &sockpri ? &sslpri : &sslaux);
#endif

    e = 0;
    s = 1;

    if (timeout >= 0) {
	tv.tv_sec = timeout;
	tv.tv_usec = 0;
	tvp = &tv;
    }
    f = fcntl(*sock, F_GETFL, 0);
    fcntl(*sock, F_SETFL, f | O_NONBLOCK);

    FD_ZERO(&fds);
    FD_SET(*sock, &fds);

#ifdef SSL_TLS
    if (*ssl) {
	for (;;) {
	    if ((s = select(*sock + 1, NULL, &fds, NULL, tvp) > 0
		 && FD_ISSET(*sock, &fds)))
		e = SSL_write(*ssl, data, strlen(data));

	    if (e > 0)
		break;

	    switch (SSL_get_error(*ssl, e)) {
	    case SSL_ERROR_WANT_READ:
	    case SSL_ERROR_WANT_WRITE:
		continue;
	    case SSL_ERROR_SYSCALL:
	    case SSL_ERROR_SSL:
		fatal(ERROR_NETWORK, "imapfilter: writing data; %s",
		      ERR_error_string(ERR_get_error(), NULL));
	    default:
		fatal(ERROR_NETWORK,
		    "imapfilter: undefined ssl error while writing data\n");
	    }
	}
    } else
#endif
	if ((s = select(*sock + 1, NULL, &fds, NULL, tvp)) > 0
	    && FD_ISSET(*sock, &fds))
	e = write(*sock, data, strlen(data));

    if (e == -1)
	fatal(ERROR_NETWORK, "imapfilter: writing data; %s",
	      strerror(errno));

    fcntl(*sock, F_SETFL, f);

    if (s == -1)
	fatal(ERROR_NETWORK, "imapfilter: waiting to write to socket; %s\n",
	      strerror(errno));
    else if (!s)
	fatal(ERROR_NETWORK,
	      "imapfilter: timeout period expired while waiting to write "
	      "data\n");

    return 0;
}
